@@ -5,12 +5,16 @@
 from __future__ import print_function
 
 import argparse
+import logging
+import os
 import os.path
 import subprocess
 import sys
 import webbrowser
 
 from dirspec.basedir import load_config_paths, save_config_path
+from iterfzf import iterfzf
+from logging_spinner import SpinnerHandler, UserWaitingFilter
 from six.moves import input
 
 from .client import (REMOTE_PATTERN, Client, ExpiredTokenIdError,
@@ -23,15 +27,17 @@ from .version import VERSION
 CONFIG_RESOURCE = 'geofront-cli'
 SERVER_CONFIG_FILENAME = 'server'
 
+WHICH_CMD = 'where' if sys.platform == 'win32' else 'which'
+
 SSH_PROGRAM = None
 try:
-    SSH_PROGRAM = subprocess.check_output(['which', 'ssh']).strip() or None
+    SSH_PROGRAM = subprocess.check_output([WHICH_CMD, 'ssh']).strip() or None
 except subprocess.CalledProcessError:
     pass
 
 SCP_PROGRAM = None
 try:
-    SCP_PROGRAM = subprocess.check_output(['which', 'scp']).strip() or None
+    SCP_PROGRAM = subprocess.check_output([WHICH_CMD, 'scp']).strip() or None
 except subprocess.CalledProcessError:
     pass
 
@@ -198,17 +204,25 @@ masterkey.add_argument(
 )
 
 
+def align_remote_list(remotes):
+    maxlength = max(map(len, remotes)) if remotes else 0
+    for alias, remote in sorted(remotes.items()):
+        if remote.endswith(':22'):
+            remote = remote[:-3]
+        yield '{0:{1}}  {2}'.format(alias, maxlength, remote)
+
+
 @subparser
 def remotes(args):
     """List available remotes."""
     client = get_client()
     remotes = client.remotes
     if args.alias:
-        for alias in remotes:
+        for alias in sorted(remotes):
             print(alias)
     else:
-        for alias, remote in remotes.items():
-            print('{0}\t{1}'.format(alias, remote))
+        for line in align_remote_list(remotes):
+            print(line)
 
 
 remotes.add_argument(
@@ -220,18 +234,25 @@ remotes.add_argument(
 
 
 @subparser
-def authorize(args):
+def authorize(args, alias=None):
     """Temporarily authorize you to access the given remote.
     A made authorization keeps alive in a minute, and then will be expired.
 
     """
     client = get_client()
-    try:
-        client.authorize(args.remote)
-    except RemoteError as e:
-        print(e, file=sys.stderr)
-        if args.debug:
-            raise
+    while True:
+        try:
+            remote = client.authorize(alias or args.remote)
+        except RemoteError as e:
+            print(e, file=sys.stderr)
+            if args.debug:
+                raise
+        except TokenIdError:
+            print('Authentication required.', file=sys.stderr)
+            authenticate.call(args)
+        else:
+            break
+    return remote
 
 
 authorize.add_argument(
@@ -301,25 +322,15 @@ colonize.add_argument('remote', help='the remote alias to colonize')
 
 
 @subparser
-def ssh(args):
+def ssh(args, alias=None):
     """SSH to the remote through Geofront's temporary authorization."""
-    while True:
-        client = get_client()
-        try:
-            remote = client.authorize(args.remote)
-        except RemoteError as e:
-            ssh.error(str(e))
-            if args.debug:
-                raise
-        except TokenIdError:
-            print('Authentication required.')
-            authenticate.call(args)
-        else:
-            break
+    remote = authorize.call(args, alias=alias)
     try:
         options = get_ssh_options(remote)
     except ValueError as e:
         ssh.error(str(e))
+    if args.jump_host:
+        options.extend(['-o', 'ProxyJump=={}'.format(args.jump_host)])
     subprocess.call([args.ssh] + options)
 
 
@@ -329,39 +340,28 @@ ssh.add_argument('remote', help='the remote alias to ssh')
 def parse_scp_path(path, args):
     """Parse remote:path format."""
     if ':' not in path:
-        return None, None, path
+        return None, path
     alias, path = path.split(':', 1)
-    while True:
-        client = get_client()
-        try:
-            remote = client.authorize(alias)
-        except RemoteError as e:
-            print(e, file=sys.stderr)
-            if args.debug:
-                raise
-            raise SystemExit(1)
-        except TokenIdError:
-            print('Authentication required.')
-            authenticate.call(args)
-        else:
-            break
-    return client, remote, path
+    remote = authorize.call(args, alias=alias)
+    return remote, path
 
 
 @subparser
 def scp(args):
     options = []
-    src_client, src_remote, src_path = parse_scp_path(args.source, args)
-    dst_client, dst_remote, dst_path = parse_scp_path(args.destination, args)
-    if src_client and dst_client:
+    src_remote, src_path = parse_scp_path(args.source, args)
+    dst_remote, dst_path = parse_scp_path(args.destination, args)
+    if src_remote and dst_remote:
         scp.error('source and destination cannot be both '
                   'remote paths at a time')
-    elif not (src_client or dst_client):
+    elif not (src_remote or dst_remote):
         scp.error('one of source and destination has to be a remote path')
     if args.ssh:
         options.extend(['-S', args.ssh])
     if args.recursive:
         options.append('-r')
+    if args.jump_host:
+        options.extend(['-o', 'ProxyJump=={}'.format(args.jump_host)])
     remote = src_remote or dst_remote
     remote_match = REMOTE_PATTERN.match(remote)
     if not remote_match:
@@ -399,7 +399,19 @@ scp.add_argument('source', help='the source path to copy')
 scp.add_argument('destination', help='the destination path')
 
 
-for p in authenticate, start, ssh, scp:
+@subparser
+def go(args):
+    """Select a remote and SSH to it at once (in interactive way)."""
+    client = get_client()
+    remotes = client.remotes
+    chosen = iterfzf(align_remote_list(remotes))
+    if chosen is None:
+        return
+    alias = chosen.split()[0]
+    ssh.call(args, alias=alias)
+
+
+for p in authenticate, authorize, start, ssh, scp, go:
     p.add_argument(
         '-O', '--no-open-browser',
         dest='open_browser',
@@ -409,8 +421,69 @@ for p in authenticate, start, ssh, scp:
     )
 
 
+for p in ssh, scp:
+    p.add_argument(
+        '-J', '--jump-host',
+        default=None,
+        help='Proxy jump host to use'
+    )
+
+
+def fix_mac_codesign():
+    """If the running Python interpreter isn't property signed on macOS
+    it's unable to get/set password using keyring from Keychain.
+
+    In such case, we need to sign the interpreter first.
+
+    https://github.com/jaraco/keyring/issues/219
+
+    """
+    global fix_mac_codesign
+    logger = logging.getLogger(__name__ + '.fix_mac_codesign')
+    p = subprocess.Popen(['codesign', '-dvvvvv', sys.executable],
+                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    stdout, stderr = p.communicate()
+
+    def prepend_lines(c, text):
+        if not isinstance(text, str):
+            text = text.decode()
+        return ''.join(c + l for l in text.splitlines(True))
+    logger.debug('codesign -dvvvvv %s:\n%s\n%s',
+                 sys.executable,
+                 prepend_lines('| ', stdout),
+                 prepend_lines('> ', stderr))
+    if b'\nSignature=' in stderr:
+        logger.debug('%s: already signed', sys.executable)
+        return
+    logger.info('%s: not signed yet; try signing...', sys.executable)
+    p = subprocess.Popen(['codesign', '-f', '-s', '-', sys.executable],
+                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    os.waitpid(p.pid, 0)
+    logger.debug('%s: signed\n%s\n%s',
+                 sys.executable,
+                 prepend_lines('| ', stdout),
+                 prepend_lines('> ', stderr))
+    logger.debug('respawn the equivalent process...')
+    raise SystemExit(subprocess.call(sys.argv))
+
+
 def main(args=None):
     args = parser.parse_args(args)
+    log_handler = logging.StreamHandler(sys.stdout)
+    log_handler.addFilter(UserWaitingFilter())
+    spinner_handler = SpinnerHandler(sys.stdout)
+    local = logging.getLogger('geofrontcli')
+    if args.debug:
+        root = logging.getLogger()
+        root.setLevel(logging.INFO)
+        root.addHandler(log_handler)
+        local.setLevel(logging.DEBUG)
+    else:
+        local.setLevel(logging.INFO)
+    local.addHandler(log_handler)
+    local.addHandler(spinner_handler)
+    if sys.platform == 'darwin':
+        fix_mac_codesign()
     if getattr(args, 'function', None):
         try:
             args.function(args)
@@ -426,3 +499,8 @@ def main(args=None):
                         'The server version is {0}.'.format(e.server_version))
     else:
         parser.print_usage()
+
+
+def main_go():
+    parser.prog = 'geofront-cli'
+    main(['go'])
