@@ -6,6 +6,7 @@ import collections
 import contextlib
 import io
 import json
+import logging
 import re
 import sys
 import uuid
@@ -14,12 +15,14 @@ from keyring import get_password, set_password
 from six import string_types
 from six.moves.urllib.error import HTTPError
 from six.moves.urllib.parse import urljoin
-from six.moves.urllib.request import Request, urlopen
+from six.moves.urllib.request import OpenerDirector, Request, build_opener
 
 from .key import PublicKey
+from .ssl import create_urllib_https_handler
 from .version import MIN_PROTOCOL_VERSION, MAX_PROTOCOL_VERSION, VERSION
 
-__all__ = ('REMOTE_PATTERN', 'Client', 'ExpiredTokenIdError',
+__all__ = ('REMOTE_PATTERN', 'BufferedResponse',
+           'Client', 'ExpiredTokenIdError',
            'MasterKeyError', 'NoTokenIdError', 'ProtocolVersionError',
            'RemoteAliasError', 'RemoteError', 'RemoteStateError',
            'TokenIdError', 'UnfinishedAuthenticationError',
@@ -51,12 +54,20 @@ class Client(object):
     #: (:class:`PublicKeyDict`) Public keys registered to Geofront server.
     public_keys = None
 
-    def __init__(self, server_url):
+    def __init__(self, server_url, opener=None):
+        self.logger = logging.getLogger(__name__ + '.Client')
         self.server_url = server_url
+        if opener is None:
+            opener = build_opener(create_urllib_https_handler())
+        elif not isinstance(opener, OpenerDirector):
+            raise TypeError('opener must be {0.__module__}.{0.__name__}, not '
+                            '{1!r}'.format(OpenerDirector, opener))
+        self.opener = opener
         self.public_keys = PublicKeyDict(self)
 
     @contextlib.contextmanager
     def request(self, method, url, data=None, headers={}):
+        logger = self.logger.getChild('request')
         if isinstance(url, tuple):
             url = './{0}/'.format('/'.join(url))
         url = urljoin(self.server_url, url)
@@ -69,8 +80,9 @@ class Client(object):
         h.update(headers)
         request = Request(url, method=method, data=data, headers=h)
         try:
-            response = urlopen(request)
+            response = self.opener.open(request)
         except HTTPError as e:
+            logger.exception(e)
             response = e
         server_version = response.headers.get('X-Geofront-Version')
         if server_version:
@@ -110,7 +122,7 @@ class Client(object):
                 raise ExpiredTokenIdError('token id seems expired')
             elif response.code == 412 and error == 'unfinished-authentication':
                 raise UnfinishedAuthenticationError(body['message'])
-            buffered = io.BytesIO(read)
+            buffered = BufferedResponse(response.code, response.headers, read)
             yield buffered
             buffered.close()
             return
@@ -170,32 +182,60 @@ class Client(object):
         addresses.
 
         """
-        path = ('tokens', self.token_id, 'remotes')
-        with self.request('GET', path) as r:
-            assert r.code == 200
-            mimetype, _ = parse_mimetype(r.headers['Content-Type'])
-            assert mimetype == 'application/json'
-            result = json.loads(r.read().decode('utf-8'))
-        fmt = '{0[user]}@{0[host]}:{0[port]}'.format
-        return dict((alias, fmt(remote)) for alias, remote in result.items())
+        logger = self.logger.getChild('remotes')
+        logger.info('Loading the list of remotes from the Geofront server...',
+                    extra={'user_waiting': True})
+        try:
+            path = ('tokens', self.token_id, 'remotes')
+            with self.request('GET', path) as r:
+                assert r.code == 200
+                mimetype, _ = parse_mimetype(r.headers['Content-Type'])
+                assert mimetype == 'application/json'
+                result = json.loads(r.read().decode('utf-8'))
+            fmt = '{0[user]}@{0[host]}:{0[port]}'.format
+            logger.info('Total %d remotes.', len(result),
+                        extra={'user_waiting': False})
+            return dict((alias, fmt(remote))
+                        for alias, remote in result.items())
+        except:
+            logger.info('Failed to fetch the list of remotes.',
+                        extra={'user_waiting': False})
+            raise
 
     def authorize(self, alias):
         """Temporarily authorize you to access the given remote ``alias``.
         A made authorization keeps alive in a minute, and then will be expired.
 
         """
-        path = ('tokens', self.token_id, 'remotes', alias)
-        with self.request('POST', path) as r:
-            mimetype, _ = parse_mimetype(r.headers['Content-Type'])
-            assert mimetype == 'application/json'
-            result = json.loads(r.read().decode('utf-8'))
-            if r.code == 404 and result.get('error') == 'not-found':
-                raise RemoteAliasError(result.get('message'))
-            elif r.code == 500 and result.get('error') == 'connection-failure':
-                raise RemoteStateError(result.get('message'))
-            assert r.code == 200
-            assert result['success'] == 'authorized'
-            return '{0[user]}@{0[host]}:{0[port]}'.format(result['remote'])
+        logger = self.logger.getChild('authorize')
+        logger.info('Letting the Geofront server to authorize you to access '
+                    'to %s...', alias, extra={'user_waiting': True})
+        try:
+            path = ('tokens', self.token_id, 'remotes', alias)
+            with self.request('POST', path) as r:
+                mimetype, _ = parse_mimetype(r.headers['Content-Type'])
+                assert mimetype == 'application/json'
+                result = json.loads(r.read().decode('utf-8'))
+                if r.code == 404 and result.get('error') == 'not-found':
+                    raise RemoteAliasError(result.get('message'))
+                elif (r.code == 500 and
+                      result.get('error') == 'connection-failure'):
+                    raise RemoteStateError(result.get('message'))
+                assert r.code == 200
+                assert result['success'] == 'authorized'
+        except TokenIdError:
+            logger.info('Authentication is required.',
+                        extra={'user_waiting': False})
+            raise
+        except:
+            logger.info('Authorization to %s has failed.', alias,
+                        extra={'user_waiting': False})
+            raise
+        else:
+            logger.info('Access to %s has authorized!  The access will be '
+                        'available only for a time.', alias,
+                        extra={'user_waiting': False})
+        return '{0[user]}@{0[host]}:{0[port]}'.format(result['remote'])
 
     def __repr__(self):
         return '{0.__module__}.{0.__name__}({1!r})'.format(
@@ -203,6 +243,18 @@ class Client(object):
         )
 
 
+class BufferedResponse(io.BytesIO):
+    """:class:`io.BytesIO` subclass that mimics some interface of
+    :class:`http.client.HTTPResponse`.
+
+    """
+
+    def __init__(self, code, headers, *args, **kwargs):
+        super(BufferedResponse, self).__init__(*args, **kwargs)
+        self.code = code
+        self.headers = headers
+
+
 class PublicKeyDict(collections.MutableMapping):
     """:class:`dict`-like object that contains public keys."""
 
